from __future__ import division
import __future__
COMPILER_FLAGS = __future__.division.compiler_flag

import sys
import os
import traceback, linecache
import re
import objc
import math
from PyObjCTools import NibClassBuilder, AppHelper
from Foundation import *
from AppKit import *
from Widgets import *
import DrawingPrimitives
import PyDETextView


VERY_LIGHT_GRAY = NSColor.blackColor().blendedColorWithFraction_ofColor_(
        0.95, NSColor.whiteColor())
DEFAULT_IMAGE_SIZE = (1000, 1000)  # XXX from user defaults?
sizeRe = re.compile("^\s*#\s*size:\s+(\d+)\s+(\d+)\s*$", re.MULTILINE)


NibClassBuilder.extractClasses("MainMenu")
NibClassBuilder.extractClasses("DrawBotDocument")


class OutputFile(object):

    def __init__(self, data, isErr=False):
        self.data = data
        self.isErr = isErr

    def write(self, data):
        if isinstance(data, str):
            try:
                data = unicode(data, "utf-8", "replace")
            except UnicodeDecodeError:
                data = "XXX " + repr(data)
        self.data.append((self.isErr, data))

class NewInterfaceError(ValueError):
    pass

# class defined in DrawBotDocument.nib
class DrawBotDocument(NibClassBuilder.AutoBaseClass):
    # the actual base class is NSDocument
    # The following outlets are added to the class:
    # drawer
    # graphicsView
    # outputView
    # textView

    path = None
    hasInterface = False
    lastInterface = None
    isNewInterface = False
    someWindow = None
    stack = 0

    def windowNibName(self):
        return "DrawBotDocument"

    def init(self):
        self = super(DrawBotDocument, self).init()
        nc = NSNotificationCenter.defaultCenter()
        nc.addObserver_selector_name_object_(self, "textFontChanged:", "PyDETextFontChanged", None)
        self.keyValueMonkey = KeyValueMonkey.alloc().init()
        self.nameTable = {
            'NSOwner': self,
            'KV': self.keyValueMonkey,
        }
        self._buildProtoNamespace()
        return self

    def _buildProtoNamespace(self):
        self.protoNamespace = namespace = {}
        namespace['defineInterface'] = self._defineInterface
        namespace['math'] = math
        for module in (math, WidgetHelpers, DrawingPrimitives):
            for name in getattr(module, '__all__', None) or dir(module):
                if not name.startswith('_'):
                    namespace[name] = getattr(module, name)

    def _newContext(self):
        self.namespace = namespace = self.protoNamespace.copy()
        return namespace
        
    def _defineInterface(self, *values):
        if self.isNewInterface:
            raise ValueError, "The interface must not change every run!"
        self.hasInterface = True
        newInterface = GSContainer(*values)
        self.isNewInterface = (newInterface != self.lastInterface)
        if self.isNewInterface:
            # replace the last interface with the new one
            # but don't do it in the callback, so raise
            self.lastInterface = newInterface
            raise NewInterfaceError()
        # In the callback, has an unchanged interface
        namespace = self.namespace
        for k, v in self.keyValueMonkey.d.iteritems():
            if not k.startswith('lblFor'):
                namespace[k] = v.objectValue()
        
    def _loadNewInterface(self):
        self.isNewInterface = False
        # release old window
        if self.someWindow:
            self.someWindow.release()
            self.someWindow = None

        # populate values from last run, if any
        # kill the dictionary
        oldDict = {}
        for k,v in self.keyValueMonkey.d.iteritems():
            if not k.startswith('lblFor'):
                oldDict[k] = v.objectValue()
        self.keyValueMonkey.d.clear()

        # load interface into self.someWindow
        filename = self.lastInterface.makeFile()
        res = NSBundle.loadGSMarkupFile_externalNameTable_withZone_localizableStringsTable_inBundle_(
            filename,
            self.nameTable,
            None,
            None,
            NSBundle.mainBundle()
        )
        os.unlink(filename)

        # populate the content of the drawer if it happens to be open
        if self.drawer and self.drawer.state() == NSDrawerOpenState:
            self.drawer.setContentView_(self.someWindow.contentView())

        # set old defaults
        for v in self.lastInterface.values:
            obj = self.keyValueMonkey.d[v.name]
            if v.default is not None:
                default = v.default
            else:
                default = oldDict.get(v.name, v.default)
            for k,v in v.other.iteritems():
                getattr(obj, 'set%s%s_' % (k[0].upper(), k[1:]))(v)
            obj.setObjectValue_(default)

        # connect up the labels
        for k,v in self.keyValueMonkey.d.iteritems():
            if k.startswith('lblFor'):
                o = self.keyValueMonkey.d[k[len('lblFor'):]]
                o.setTarget_(v)
                o.setAction_('takeObjectValueFrom:')
                v.takeObjectValueFrom_(o)

    def __del__(self):
        nc = NSNotificationCenter.defaultCenter()
        nc.removeObserver_name_object_(self, "PyDETextFontChanged", None)
        # text view has a couple of circular refs, it can let go of them now
        self.textView._cleanup()

    def textFontChanged_(self, notification):
        font = PyDETextView.getBasicTextAttributes()[NSFontAttributeName]
        self.outputView.setFont_(font)

    def readFromFile_ofType_(self, path, tp):
        if self.textView is None:
            # we're not yet fully loaded
            self.path = path
        else:
            # "revert"
            self.readFromUTF8(path)
        return True

    def drawerShouldOpen_(self, aNotification):
        return self.hasInterface

    def drawerWillOpen_(self, aNotification):
        self.drawer.setContentView_(self.someWindow.contentView())

    def writeToFile_ofType_(self, path, tp):
        f = file(path, "w")
        text = self.textView.string()
        f.write(text.encode("utf8"))
        f.close()
        return True

    def windowControllerDidLoadNib_(self, controller):
        if self.path:
            self.readFromUTF8(self.path)
        font = PyDETextView.getBasicTextAttributes()[NSFontAttributeName]
        self.outputView.setFont_(font)
        self.windowControllers()[0].setWindowFrameAutosaveName_("DrawBotDocumentWindow")

    def readFromUTF8(self, path):
        f = file(path)
        text = f.read().decode("utf8")
        f.close()
        self.textView.setString_(text)
        self.textView.usesTabs = "\t" in text

    def runScript_(self, sender=None):
        # We have to turn everything inside out a couple of times. The only
        # way I have found to create a vector-based PDF image, is by calling
        # dataWithPDFInsideRect_ on a view that draws vectors. We use the
        # graphicsView to do that: we give it a callback, which will be
        # invoked by its drawRect_ method. So we execute the script from
        # within the call to dataWithPDFInsideRect_/drawRect_.

        hadInterface = self.hasInterface
        self.hasInterface = False
        drawerWasOpen = self.drawer and self.drawer.state() == NSDrawerOpenState

        m = sizeRe.search(self.textView.string())
        if m:
            size = tuple(map(float, m.groups()))
        else:
            size = DEFAULT_IMAGE_SIZE
        self.graphicsView.drawCallback = self._runScriptInternal
        try:
            pdfData = self.graphicsView.dataWithPDFInsideRect_(((0, 0), size))
        finally:
            self.graphicsView.drawCallback = None
        self.graphicsView.setPDFData(pdfData, size)
        # Must do update stdout/stderr output totally separately, otherwise we
        # may get text turds in our graphic.
        self._flushOutput()

        if hadInterface and not self.hasInterface:
            # There was an interface, now there's not. Clean up!
            if drawerWasOpen:
                self.toggleDrawer_(self)
            if self.someWindow:
                self.someWindow.release()
                self.someWindow = None
            self.lastInterface = None
            self.keyValueMonkey.d.clear()
        if self.isNewInterface and self.stack == 0:
            self._loadNewInterface()
            # stack is to prevent infinite recursion
            self.stack = 1
            self.runScript_(sender)
            self.stack = 0
        if self.hasInterface and not hadInterface:
            if not drawerWasOpen:
                self.toggleDrawer_(self)

    def _runScriptInternal(self):
        # Note that this happens in a callback while drawing!
        fileName = self.fileName()
        if not fileName:
            curDir = os.getenv("HOME")
            fileName = "<untitled>"
        else:
            curDir = os.path.dirname(fileName)
        source = self.textView.string()

        linecache.clearcache()
        linecache.cache[fileName] = len(source), 0, source.splitlines(True), fileName
        save = sys.stdout, sys.stderr
        saveDir = os.getcwd()
        saveArgv = sys.argv
        sys.argv = [fileName]
        os.chdir(curDir)
        sys.path.insert(0, curDir)
        self.output = []
        sys.stdout = OutputFile(self.output, False)
        sys.stderr = OutputFile(self.output, True)
        try:
            try:
                code = compile(source + "\n\n", fileName, "exec", COMPILER_FLAGS)
            except:
                traceback.print_exc(0)
            else:
                try:
                    exec code in self._newContext()
                except NewInterfaceError:
                    # defineInterface used, created a new interface
                    pass
                except:
                    etype, value, tb = sys.exc_info()
                    if tb.tb_next is not None:
                        tb = tb.tb_next  # skip the frame doing the exec
                    traceback.print_exception(etype, value, tb)
                    etype = value = tb = None
        finally:
            sys.stdout, sys.stderr = save
            os.chdir(saveDir)
            sys.path.remove(curDir)
            sys.argv = saveArgv

    def _flushOutput(self):
        outAttrs = PyDETextView.getBasicTextAttributes()
        errAttrs = outAttrs.copy()
        # XXX err color from user defaults...
        errAttrs[NSForegroundColorAttributeName] = NSColor.redColor()

        outputView = self.outputView
        outputView.setSelectedRange_((outputView.textStorage().length(), 0))
        lastErr = None
        for isErr, data in self.output:
            if isErr != lastErr:
                attrs = [outAttrs, errAttrs][isErr]
                outputView.setTypingAttributes_(attrs)
                lastErr = isErr
            outputView.insertText_(data)
        del self.output

    def toggleDrawer_(self, sender):
        self.drawer.toggle_(sender)

    def copyImageAsPDF_(self, sender):
        pboard = NSPasteboard.generalPasteboard()
        # graphicsView implements the pboard delegate method to provide the data
        pboard.declareTypes_owner_([NSPDFPboardType], self.graphicsView)

    def saveDocumentAsPDF_(self, sender):
        savePanel = NSSavePanel.savePanel()
        path = self.fileName()
        if path:
            dirName, fileName = os.path.split(path)
            fileName, ext = os.path.splitext(fileName)
            fileName += ".pdf"
        else:
            dirName, fileName = None, "Untitled.pdf"
        savePanel.beginSheetForDirectory_file_modalForWindow_modalDelegate_didEndSelector_contextInfo_(
            dirName, fileName, NSApp().mainWindow(), self,
            "savePanelDidEnd:returnCode:contextInfo:", 0)

    def savePanelDidEnd_returnCode_contextInfo_(self, panel, returnCode, context):
        if returnCode:
            pdfData = self.graphicsView.pdfData
            pdfData.writeToFile_atomically_(panel.filename(), False)

    savePanelDidEnd_returnCode_contextInfo_ = objc.selector(savePanelDidEnd_returnCode_contextInfo_,
            signature="v@:@ii")

    def printDocument_(self, sender):
        op = NSPrintOperation.printOperationWithView_printInfo_(self.graphicsView, self.printInfo())
        op.runOperationModalForWindow_delegate_didRunSelector_contextInfo_(
            NSApp().mainWindow(), self, "printOperationDidRun:success:contextInfo:",
            0)

    def printOperationDidRun_success_contextInfo_(self, op, success, info):
        if success:
            self.setPrintInfo_(op.printInfo())

    printOperationDidRun_success_contextInfo_ = objc.selector(printOperationDidRun_success_contextInfo_,
            signature="v@:@ci")

    def validateMenuItem_(self, menuItem):
        if menuItem.action() in ("saveDocumentAsPDF:", "copyImageAsPDF:"):
            return self.graphicsView.image is not None
        return True


# class defined in DrawBotDocument.nib
class DrawBotGraphicsView(NibClassBuilder.AutoBaseClass):
    # the actual base class is NSView

    def awakeFromNib(self):
        self.image = None
        self.pdfRep = None
        self.drawCallback = None
        self.setFrameSize_(DEFAULT_IMAGE_SIZE)
        self.superview().setBackgroundColor_(VERY_LIGHT_GRAY)

    def setPDFData(self, pdfData, size=None):
        assert self.drawCallback is None
        self.pdfRep = NSPDFImageRep.imageRepWithData_(pdfData)
        self.pdfData = pdfData
        if size is None:
            size = self.frame()[1]
        else:
            self.setFrameSize_(size)
        self.image = NSImage.alloc().initWithSize_(size)
        self.image.addRepresentation_(self.pdfRep)
        self.setNeedsDisplay_(True)

    def isOpaque(self):
        return True

    def drawRect_(self, rect):
        if self.drawCallback is not None:
            try:
                self.drawCallback()
            except:
                print "A Bug in DrawBot:"
                traceback.print_exc()
        elif NSGraphicsContext.currentContext().isDrawingToScreen():
            NSColor.whiteColor().set()
            NSRectFill(self.bounds())
            if self.image is not None:
                self.image.drawAtPoint_fromRect_operation_fraction_((0, 0),
                    self.frame(), NSCompositeSourceOver, 1.0)
        else:
            if self.pdfRep is not None:
                self.pdfRep.draw()

    def mouseDown_(self, event):
        if self.image is None:
            return
        self.startDrag = True

    def mouseDragged_(self, event):
        if self.image is None:
            return
        if not self.startDrag:
            return
        self.startDrag = False
        pboard = NSPasteboard.pasteboardWithName_(NSDragPboard)
        pboard.declareTypes_owner_([NSPDFPboardType], self)
        w, h = self.image.size()
        srcRect = ((0, 0), (w, h))
        if w > 400 or h > 400:
            if w > h:
                scale = 400.0 / w
            else:
                scale = 400.0 / h
            dstRect = ((0, 0), (scale * w, scale * h))
            x, y = self.convertPoint_fromView_(event.locationInWindow(), None)
            offset = x * (1 - scale), y * (1 - scale)
        else:
            dstRect = srcRect
            offset = (0, 0)
        drawImage = NSImage.alloc().initWithSize_(dstRect[1])
        drawImage.lockFocus()
        self.image.drawInRect_fromRect_operation_fraction_(
            dstRect, srcRect, NSCompositeSourceOver, 0.75)
        drawImage.unlockFocus()
        self.dragImage_at_offset_event_pasteboard_source_slideBack_(
            drawImage, offset, (0, 0), event, pboard, self, True)

    # pasteboard delegate method
    def pasteboard_provideDataForType_(self, pboard, typ):
        if typ == NSPDFPboardType:
            pboard.setData_forType_(self.pdfData, NSPDFPboardType)

    # These two methods, together with the logic in mouseDown_ and mouseDragged_,
    # are needed to get the proper drag behavior whenever our app is in the
    # background. Ie. it will remain in the background when you drag from it,
    # but will come to the front for a non-drag click.
    def acceptsFirstMouse_(self, event):
        return True

    def shouldDelayWindowOrderingForEvent_(self, event):
        return True


class DrawBotAppDelegate(NibClassBuilder.AutoBaseClass):

    def awakeFromNib(self):
        self._prefsController = None
        NSColor.setIgnoresAlpha_(False)

    def showPreferencesPanel_(self, sender):
        if self._prefsController is None:
            from DrawBotPreferencesPanel import DrawBotPreferencesController
            self._prefsController = DrawBotPreferencesController.alloc().init()
        self._prefsController.showWindow_(sender)


if __name__ == "__main__":
    AppHelper.runEventLoop()
