import tempfile
from Foundation import *
from AppKit import *
from Renaissance import *

__all__ = ['KeyValueMonkey', 'GSContainer', 'WidgetHelpers']

class KeyValueMonkey(NSObject):
    """This is basically like a dictionary that supports
    key-value coding"""
    def init(self):
        self = super(KeyValueMonkey, self).init()
        self.d = {}
        return self

    def handleQueryWithUnboundKey_(self, key):
        return self.d[key]

    def handleTakeValue_forUnboundKey_(self, value, key):
        self.d[key] = value

    def __getitem__(self, item):
        return self.d[item]

    def __setitem__(self, item, value):
        self.d[item] = value

    def __iter__(self):
        return iter(self.d)

# XML template for gsmarkup
GSTEMPLATE = """<?xml version="1.0"?>
<gsmarkup>
  <objects>
    <window visible="no" id="window">
      <box title="Variables">
      <vbox>
            %(objects)s
      </vbox>
      </box>
    </window>
  </objects>
  <connectors>
        <outlet source="#NSOwner" target="#window" key="someWindow" />
        %(connectors)s
  </connectors>
</gsmarkup>"""

# Generic template for a label
LABELTEMPLATE = """<label halign="left">%(name)s</label>"""
# Generic template for a NSView subclass
VIEWTEMPLATE  = """<view halign="left" instanceOf="%(classname)s" width="%(width)s" height="%(height)s" id="%(name)s" />"""
# Generic template for a connector
CONNECTORTEMPLATE = """<outlet source="#KV" target="#%(name)s" key="%(name)s" />"""

class GSContainer(object):
    """A container for widgets that can be printed as XML.  Can be
    compared to other GSContainers for equality to determine if the
    layout has been changed."""
    def __init__(self, *values):
        self.values = list(values)
    
    def makeFile(self):
        """Create the gsmarkup file, does not delete it.  Receiver is
        responsible for unlinking"""
        fd, name = tempfile.mkstemp(suffix='.gsmarkup')
        fileobj = os.fdopen(fd, 'w+b')
        fileobj.write(self.asXML())
        fileobj.close()
        return name

    def __eq__(self, other):
        return isinstance(other, GSContainer) and self.values == other.values

    def __ne__(self, other):
        return not (self == other)

    def asXML(self):
        """Return a gsmarkup document as a string"""
        objects = []
        connectors = []
        for obj in self.values:
            objects.append(obj.objectXML())
            connectors.append(obj.connectorXML())
        xml = GSTEMPLATE % dict(objects='\n'.join(objects), connectors='\n'.join(connectors))
        return xml
        
class GSThing(object):
    def __init__(self, name, classname, width, height, default=None, **other):
        self.name = name
        self.classname = classname
        self.width = width
        self.height = height
        self.default = default
        self.other = other

    def objectXML(self):
        """Returns the contents that belong in <objects>"""
        return ('%s\n%s' % (LABELTEMPLATE, VIEWTEMPLATE)) % self.__dict__

    def connectorXML(self):
        """Returns the contents that belong in <connectors>"""
        return CONNECTORTEMPLATE % self.__dict__

    def __eq__(self, other):
        return (
            isinstance(other, GSThing) and 
            self.name == other.name and
            self.classname == other.classname and
            self.width == other.width and
            self.height == other.height and
            self.default == other.default and
            self.other == other.other
        )

    def __ne__(self, other):
        return not (self == other)

class WidgetHelpers(object):
    """This is a convenient place to put classes that get
    imported into the DrawBot code namespace"""
    class Widget(GSThing):
        """Generic NSView subclass widget"""
        pass

    class ColorWell(GSThing):
        """A color well"""
        def __init__(self, name, width=25, height=25, default=None):
            """Takes a RGB[A] tuple/list, NSColor, or Color for default"""
            if isinstance(default, (tuple, list)):
                default = list(default)
                if len(default) == 3:
                    default.append(1.0)
                default = NSColor.colorWithCalibratedRed_green_blue_alpha_(*default)
            elif isinstance(default, basestring):
                default = getattr(NSColor, str(default.lower())+'Color')()
            GSThing.__init__(self, name=name, classname='NSColorWellFix',
                width=width, height=height, default=default)

    class Slider(GSThing):
        """A horizontal NSSlider/label pair"""
        def __init__(self, name, minValue=0.0, maxValue=1.0, tickMarks=10,
                clamp=False, width=150, height=25, default=None):
            """clamp means that it only allows tickMark values
            tickMarks is the number of tick marks not including zero
            Takes a number for default
            """
            GSThing.__init__(self, name=name, classname='NSSlider', width=width,
                height=height, default=default, minValue=minValue,
                maxValue=maxValue, numberOfTickMarks=tickMarks+1,
                allowsTickMarkValuesOnly=clamp)
   
        def objectXML(self):
            LABELTEMPLATE = """<hbox halign="left"><label halign="left">%(name)s:</label><label halign="wexpand" id="lblFor%(name)s" /></hbox>"""
            VIEWTEMPLATE  = """<view halign="left" instanceOf="%(classname)s" width="%(width)s" height="%(height)s" id="%(name)s" />"""
            return ('%s\n%s' % (LABELTEMPLATE, VIEWTEMPLATE)) % self.__dict__

        def connectorXML(self):
            CONNECTORTEMPLATE = '\n'.join((
                """<outlet source="#KV" target="#%(name)s" key="%(name)s" />""",
                """<outlet source="#KV" target="#lblFor%(name)s" key="lblFor%(name)s" />""",
            ))
            return CONNECTORTEMPLATE % self.__dict__

    class Checkbox(GSThing):
        """A checkbox/label pair"""
        def __init__(self, name, default=None):
            """Takes a bool for default"""
            GSThing.__init__(self, name=name, classname='NSSwitchButton', width=None, height=None, default=default)

        def objectXML(self):
            LABELVIEWTEMPLATE = """<button title="%(name)s" id="%(name)s" type="switch" halign="wexpand" />"""
            return LABELVIEWTEMPLATE % self.__dict__
            
 
class NSColorWellFix(NSColorWell):
    """This is a NSColorWell that tries to hold on to 
    FirstResponder while you're using it.  It's not 100%,
    because FirstResponder changes if don't close the
    color panel right after choosing the color or whatever.

    It's functional enough to use, though."""

    # ivars
    _oldResponder = None
    _sendsFirstResponderSetColorMessage = None

    # void
    def awakeFromGSMarkup(self):
        self.setSendsFirstResponderSetColorMessage_(False)
        self._oldResponder = None

    # void
    def awakeFromNib(self):
        self.setSendsFirstResponderSetColorMessage_(False)
        self._oldResponder = None

    # void
    def setSendsFirstResponderSetColorMessage_(self, aBool):
        self._sendsFirstResponderSetColorMessage = aBool

    # BOOL
    def sendsFirstResponderSetColorMessage(self):
        return self._sendsFirstResponderSetColorMessage

    def objectValue(self):
        return self.color()

    def setObjectValue_(self, obj):
        if obj:
            self.setColor_(obj)

    # void
    def mouseDown_(self, theEvent):
        if (self.sendsFirstResponderSetColorMessage() or self._oldResponder is not None):
            super(NSColorWellFix, self).mouseDown_(theEvent)
            return
        self._oldResponder = self.window().firstResponder()
        self.window().makeFirstResponder_(self)
        super(NSColorWellFix, self).mouseDown_(theEvent)
        NSNotificationCenter.defaultCenter().addObserver_selector_name_object_(
            self,
            'colorPanelNotification:',
            None,
            NSColorPanel.sharedColorPanel())
        return

    # void
    def colorPanelNotification_(self, notification):
        if notification.name() == NSWindowWillCloseNotification:
            self.window().makeFirstResponder_(self._oldResponder)
            self._oldResponder = None
            NSNotificationCenter.defaultCenter().removeObserver_(self)
